from __future__ import annotations

from sqlalchemy import types

from subtypes import DateTime, Date


class BitLiteral(types.TypeDecorator):
    impl = types.DateTime

    def process_literal_param(self, value, dialect):
        return str(int(value))


class SubtypesDateTime(types.TypeDecorator):
    impl = types.DateTime
    string = types.String()

    def process_bind_param(self, value, dialect):
        return None if value is None else DateTime.from_datelike(value).to_isoformat()

    def process_literal_param(self, value, dialect):
        return None if value is None else self.string.literal_processor(dialect)(DateTime.from_datelike(value).to_isoformat())

    def process_result_value(self, value, dialect):
        return None if value is None else DateTime.from_datelike(value)


class SubtypesDate(types.TypeDecorator):
    impl = types.Date
    string = types.String()

    def process_bind_param(self, value, dialect):
        return None if value is None else Date.from_datelike(value).to_isoformat()

    def process_literal_param(self, value, dialect):
        return None if value is None else self.string.literal_processor(dialect)(Date.from_datelike(value).to_isoformat())

    def process_result_value(self, value, dialect):
        return None if value is None else Date.from_datelike(value)



from __future__ import annotations

from typing import Any

from django import db

from subtypes import Dict

from .config import SqlConfig
from .database import DjangoDatabase, DjangoApps
from .model import SqlModel

from sqlhandler.utils import Url


class DjangoSql(SqlConfig.Sql):
    class Settings(SqlConfig.Sql.Settings):
        eager_reflection = cache_metadata = True
        reflect_views = False

    class Constructors(SqlConfig.Sql.Constructors):
        ReflectedModel, Database = SqlModel, DjangoDatabase

    @property
    def django(self) -> DjangoApps:
        return self.database.django

    def _create_url(self, connection: str, **kwargs: Any) -> Url:
        detail = Dict(db.connections.databases[connection])
        drivername = SqlConfig.settings.ENGINES[detail.ENGINE.rpartition(".")[-1]]
        return Url(drivername=drivername, database=detail.NAME, username=detail.USER or None, password=detail.PASSWORD or None, host=detail.HOST or None, port=detail.PORT or None)



from __future__ import annotations

from typing import Any

import sqlalchemy as alch
from sqlalchemy.orm import backref

from subtypes import Frame

from sqlhandler.utils import literal_statement


class Session(alch.orm.Session):
    """Custom subclass of sqlalchemy.orm.Session granting access to a custom Query class through the '.query()' method."""

    def query(self, *entities: Any) -> Query:
        return super().query(*entities)

    def execute(self, *args: Any, autocommit: bool = False, **kwargs: Any) -> alch.engine.ResultProxy:
        """Execute an valid object against this Session. If 'autocommit=True' is passed, the transaction will be commited if the statement completes without errors."""
        res = super().execute(*args, **kwargs)
        if autocommit:
            self.commit()
        return res


class Query(alch.orm.Query):
    """Custom subclass of sqlalchemy.orm.Query with additional useful methods and aliases for existing methods."""

    def __repr__(self) -> str:
        return f"{type(self).__name__}(\n{(str(self))}\n)"

    def __str__(self) -> str:
        return literal_statement(self)

    def frame(self, labels: bool = False) -> Frame:
        """Execute the query and return the result as a pandas DataFrame."""
        return self.session.bind.sql.query_to_frame(self, labels=labels)

    def vector(self) -> list:
        """Transpose all records in a single column into a list. If the query returns more than one column, this will raise a RuntimeError."""
        vals = self.all()
        if all([len(row) == 1 for row in vals]):
            return [row[0] for row in vals]
        else:
            raise RuntimeError("Multiple columns selected. Expected exactly one value per row, got multiple.")

    def from_(self, *from_obj: Any) -> Query:
        """Simple alias for the 'select_from' method. See that method's docstring for documentation."""
        return self.select_from(*from_obj)

    def where(self, *criterion: Any) -> Query:
        """Simple alias for the 'filter' method. See that method's docstring for documentation."""
        return self.filter(*criterion)

    def update(self, values: Any, synchronize_session: Any = "fetch", update_args: dict = None) -> int:
        """Simple alias for the '.update()' method, with the default 'synchronize_session' argument set to 'fetch', rather than 'evaluate'. Check that method for documentation."""
        return super().update(values, synchronize_session=synchronize_session)

    def delete(self, synchronize_session: Any = "fetch") -> int:
        """Simple alias for the '.delete()' method, with the default 'synchronize_session' argument set to 'fetch', rather than 'evaluate'. Check that method for documentation."""
        return super().delete(synchronize_session=synchronize_session)

    def subquery(self, name: str = None, with_labels: bool = False, reduce_columns: bool = False):
        sub = super().subquery(name=name, with_labels=with_labels, reduce_columns=reduce_columns)
        for col in sub.c:
            setattr(sub, col.name, col)
        return sub




from __future__ import annotations

from typing import Any, TYPE_CHECKING

import sqlalchemy as alch
from sqlalchemy.schema import CreateTable
from sqlalchemy.orm.attributes import InstrumentedAttribute

if TYPE_CHECKING:
    from .model import ModelMeta


class CreateTableAccessor:
    def __init__(self, model_cls: ModelMeta) -> None:
        self.model_cls = model_cls

    def __repr__(self) -> str:
        return str(CreateTable(self.model_cls.__table__)).strip()

    def __call__(self) -> str:
        return self.model_cls.metadata.sql.database.create_table(self.model_cls)


class ForeignKey(alch.ForeignKey):
    def __init__(self, column: Any, *args: Any, **kwargs: Any) -> None:
        super().__init__(column=column.comparator.table.c[column.comparator.key] if isinstance(column, InstrumentedAttribute) else column, *args, **kwargs)


def absolute_namespace(bases: tuple, namespace: dict) -> dict:
    abs_ns = {}
    for immediate_base in reversed(bases):
        for hierarchical_base in reversed(immediate_base.mro()):
            abs_ns.update(vars(hierarchical_base))

    abs_ns.update(namespace)

    return abs_ns



from __future__ import annotations

from maybe import Maybe
from subtypes import ValueEnum
import iotools
import sqlhandler

from sqlalchemy.engine.url import URL
from sqlalchemy import create_engine


class Dialect(ValueEnum):
    """Enum of known dialect drivers."""
    MS_SQL, MY_SQL, SQLITE, POSTGRESQL, ORACLE = "mssql", "mysql", "sqlite", "postgresql+psycopg2", "oracle"


class Url(URL):
    """Class representing a sql connection URL."""

    def __init__(self, drivername: str = None, database: str = None, username: str = None, password: str = None, host: str = None, port: str = None, query: dict = None) -> None:
        super().__init__(drivername=drivername, username=Maybe(username).else_(""), password=password, host=host, port=port, database=database, query=query)


class Config(iotools.Config):
    """A config class granting access to an os-specific appdata directory for use by this library."""
    Dialect = Dialect
    name = sqlhandler.__name__
    default = {"default_connection": "memory", "connections": {"memory": {"drivername": "sqlite", "default_database": None, "username": None, "password": None, "host": "", "port": None, "query": None}}}

    def add_connection(self, connection: str, drivername: str, default_database: str, username: str = None, password: str = None, host: str = None, port: str = None, query: dict = None, is_default: bool = False) -> None:
        """Add a new connection with the given arguments."""
        self.data.connections[connection] = dict(drivername=drivername, default_database=default_database, username=username, password=password, host=host, port=port, query=query)
        if is_default:
            self.set_default_connection(connection=connection)

    def add_mssql_connection_with_integrated_security(self, connection: str, default_database: str, host: str, is_default: bool = False):
        """Add a SQL server connection that will use Windows integrated security."""
        self.add_connection(connection=connection, drivername=Dialect.MS_SQL, default_database=default_database, host=host, query={"driver": "SQL+Server"}, is_default=is_default)

    def set_default_connection(self, connection: str) -> None:
        """Set the connection that will be used by default."""
        if connection in self.data.connections:
            self.data.default_connection = connection
        else:
            raise ValueError(f"Connection {connection} is not one of the currently registered connections: {', '.join(self.data.connections)}. Use {type(self).__name__}.add_connection() first.")

    def generate_url(self, connection: str = None, database: str = None) -> Url:
        """Generate a sql connection URL from the current config with optional overrides passed as arguments."""
        if connection is None or connection in self.data.connections:
            settings = self.data.connections[Maybe(connection).else_(self.data.default_connection)]
            database = Maybe(database).else_(settings.default_database)
            return Url(drivername=settings.drivername, database=database, username=settings.username, password=settings.password, host=settings.host, port=settings.port, query=Maybe(settings.query).else_(None))
        else:
            return create_engine(connection).url



from __future__ import annotations

from typing import Any, Callable, Optional, TYPE_CHECKING

from sqlalchemy import Column
from sqlalchemy import types
from sqlalchemy.orm import backref, relationship

from subtypes import Str, Dict, Enum

from .misc import ForeignKey, absolute_namespace

if TYPE_CHECKING:
    from .model import BaseModel, Table

# TODO: Find way to implement ONE_TO_MANY relationship by extending the previous model with a foreign key after the fact


class Relationship:
    class Settings:
        casing, fk_suffix, association_table_suffix = Str.Case.SNAKE, "id", "mapping"
        default_backref_kwargs = {
            "cascade": "all"
        }

    class Kind(Enum):
        ONE_TO_ONE, MANY_TO_ONE, MANY_TO_MANY = "one_to_one", "many_to_one", "many_to_many"

    class One:
        @classmethod
        def to_one(cls, target: BaseModel, backref_name: str = None, **backref_kwargs: Any) -> Relationship:
            return Relationship(target=target, kind=Relationship.Kind.ONE_TO_ONE, backref_name=backref_name, **backref_kwargs)

    class Many:
        @classmethod
        def to_one(cls, target: BaseModel, backref_name: str = None, **backref_kwargs: Any) -> Relationship:
            return Relationship(target=target, kind=Relationship.Kind.MANY_TO_ONE, backref_name=backref_name, **backref_kwargs)

        @classmethod
        def to_many(cls, target: BaseModel, backref_name: str = None, association: str = None, **backref_kwargs: Any) -> Relationship:
            return Relationship(target=target, kind=Relationship.Kind.MANY_TO_MANY, backref_name=backref_name, association=association, **backref_kwargs)

    class _TargetEntity:
        def __init__(self, model: BaseModel, rel: Relationship) -> None:
            self.relationship, self.model, self.name = rel, model, model.__table__.name
            self.pk, = list(self.model.__table__.primary_key)
            self.fk = self.relationship._casing(f"{self.name}_{rel.settings.fk_suffix}")

        def __repr__(self) -> str:
            return f"{type(self).__name__}({', '.join([f'{attr}={repr(val)}' for attr, val in self.__dict__.items() if not attr.startswith('_')])})"

    class _FutureEntity:
        def __init__(self, table_name: str, bases: tuple, namespace: dict, rel: Relationship) -> None:
            self.relationship, self.name, self.bases, self.namespace = rel, table_name, bases, namespace
            self.plural = str(Str(self.name).case.plural())

            pk_key, = [key for key, val in absolute_namespace(bases=bases, namespace=namespace).items() if isinstance(val, Column) and val.primary_key]
            self.pk = f"{self.name}.{pk_key}"

        def __repr__(self) -> str:
            return f"{type(self).__name__}({', '.join([f'{attr}={repr(val)}' for attr, val in self.__dict__.items() if not attr.startswith('_')])})"

    def __init__(self, target: BaseModel, kind: Relationship.Kind, backref_name: str = None, association: str = None, **backref_kwargs: Any) -> None:
        self.settings = self.Settings()

        self.target, self.kind, self.backref_name, self.association = Relationship._TargetEntity(model=target, rel=self), kind, backref_name, association
        self.backref_kwargs = Dict({**self.settings.default_backref_kwargs, **backref_kwargs})

        self.this: Optional[Relationship._FutureEntity] = None
        self.attribute: Optional[str] = None

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join([f'{attr}={repr(val)}' for attr, val in self.__dict__.items() if not attr.startswith('_')])})"

    def build(self, table_name: str, bases: tuple, namespace: dict, attribute: str) -> None:
        self.this, self.attribute = Relationship._FutureEntity(table_name=table_name, bases=bases, namespace=namespace, rel=self), attribute
        self._build_fk_columns()
        self._build_relationship()

    def _build_fk_columns(self) -> None:
        if self.kind == self.Kind.MANY_TO_ONE:
            self.this.namespace[self.target.fk] = Column(types.Integer, ForeignKey(self.target.pk))
        elif self.kind == self.Kind.ONE_TO_ONE:
            self.this.namespace[self.target.fk] = Column(types.Integer, ForeignKey(self.target.pk), unique=True)
        else:
            self.Kind(self.kind)

    def _build_relationship(self) -> None:
        if self.backref_name is not None:
            backref_name = self.backref_name
        else:
            if self.kind == self.Kind.ONE_TO_ONE:
                backref_name = self.this.name
            elif self.kind in (self.Kind.MANY_TO_ONE, self.Kind.MANY_TO_MANY):
                backref_name = self.this.plural
            else:
                self.Kind(self.kind)
                backref_name = None

        if self.kind == Relationship.Kind.ONE_TO_ONE:
            self.backref_kwargs.uselist = False

        secondary = self._build_association_table() if self.kind == Relationship.Kind.MANY_TO_MANY else None

        self.this.namespace[self.attribute] = relationship(self.target.model, secondary=secondary, backref=backref(name=backref_name, **self.backref_kwargs))

    # noinspection PyArgumentList
    def _build_association_table(self) -> Callable[[], Table]:
        if self.association is not None:
            table = self.association
        else:
            from .model import Table

            name = self._casing(f"association_{self.this.name}_{self.target.name}")
            this_col = Column(self._casing(f"{self.this.name}_{self.settings.fk_suffix}"), types.Integer, ForeignKey(self.this.pk))
            target_col = Column(self._casing(f"{self.target.name}_{self.settings.fk_suffix}"), types.Integer, ForeignKey(self.target.pk))
            table = Table(name, self.target.model.metadata, Column(self._casing("id"), types.Integer, primary_key=True), this_col, target_col)

        self.this.namespace[self._casing(f"{self.target.name}_{self.settings.association_table_suffix}")] = table
        setattr(self.target.model, self._casing(f"{self.this.name}_{self.settings.association_table_suffix}"), table)

        def _defer_create_table() -> Table:
            table.create()
            return table

        return _defer_create_table

    def _casing(self, text: str) -> str:
        return str(Str(text).case.from_enum(self.settings.casing))



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import sqlalchemy as alch

from sqlhandler.database.name import SchemaName

if TYPE_CHECKING:
    from sqlhandler import Sql


class Metadata(alch.MetaData):
    def __init__(self, sql: Sql = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.sql = sql

    def __repr__(self) -> str:
        return f"{type(self).__name__}(tables={len(self.tables)})"

    def schema_subset(self, schema: SchemaName) -> Metadata:
        meta = type(self)(sql=self.sql, bind=self.sql.engine)
        meta.tables = type(self.tables)({name: table for name, table in self.tables.items() if table.schema == schema})
        return meta


class NullRegistry(dict):
    def __setitem__(self, key: Any, val: Any) -> None:
        pass



from __future__ import annotations

from typing import Any

from iotools import PrintLog


class SqlLog(PrintLog):
    """A logger designed to leave behind properly formatted SQL scripts from any SQL statements executed using the '.resolve()' method of the custom expression classes."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.to_file = False

    def write_sql(self, text: str, add_newlines: int = 2) -> None:
        """Write a SQL statement to the log if the log is active. Always write it to the stream."""
        self.write(text=text, to_stream=True, to_file=False, add_newlines=add_newlines)

        if self.active:
            self.write(text=text, to_stream=False, to_file=True, add_newlines=add_newlines)

    def write_comment(self, text: str, single_line_comment_cutoff: int = 5, add_newlines: int = 2) -> None:
        """Write a SQL comment to the log if the log is active in either short or long-form depending on the 'single_line_comment_cutoff'. Always write it to the stream."""
        self.write(text=text, to_stream=True, to_file=False, add_newlines=add_newlines)

        if self.active:
            if text.strip().count("\n") <= single_line_comment_cutoff:
                text = "-- " + text.strip().replace("\n", "\n-- ")
            else:
                text = "/*\n" + text.strip() + "\n*/"

            self.write(text=text, to_stream=False, to_file=True, add_newlines=add_newlines)

    @classmethod
    def from_details(cls, log_name: str, log_dir: str = None, active: bool = True, file_extension: str = "sql") -> SqlLog:
        """Create a SqlLog from several arguments."""
        return super().from_details(log_name=log_name, file_extension=file_extension, log_dir=log_dir, active=active)



from setuptools import setup, find_packages
from os import path

__version__ = "0.3.10"

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open(path.join(here, "requirements.txt"), encoding="utf-8") as f:
    dependencies = [line for line in f if line]

setup(
    name="sqlhandler",
    version=__version__,
    description="Sql handler class controlling sqlalchemy resources and granting access to sqlalchemy object subclasses with additional convenience methods.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/matthewgdv/sqlhandler",
    license="MIT",
    classifiers=[
      "Development Status :: 3 - Alpha",
      "Intended Audience :: Developers",
      "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages(exclude=["tests*"]),
    install_requires=dependencies,
    author="Matt GdV",
    author_email="matthewgdv@gmail.com"
)
